@@ -47,14 +47,17 @@
 #include <gdbus.h>
 
 #include "logging.h"
+#include "error.h"
 #include "uinput.h"
 #include "adapter.h"
+#include "../src/device.h"
 #include "device.h"
 #include "manager.h"
 #include "avdtp.h"
 #include "control.h"
 #include "sdpd.h"
 #include "glib-helper.h"
+#include "btio.h"
 #include "dbus-common.h"
 
 #define AVCTP_PSM 23
@@ -72,7 +75,9 @@
 /* ctype entries */
 #define CTYPE_CONTROL		0x0
 #define CTYPE_STATUS		0x1
+#define CTYPE_NOT_IMPLEMENTED	0x8
 #define CTYPE_ACCEPTED		0x9
+#define CTYPE_REJECTED		0xA
 #define CTYPE_STABLE		0xC
 
 /* opcodes */
@@ -84,28 +89,22 @@
 #define SUBUNIT_PANEL		0x09
 
 /* operands in passthrough commands */
-#define VOLUP_OP		0x41
-#define VOLDOWN_OP		0x42
+#define VOL_UP_OP		0x41
+#define VOL_DOWN_OP		0x42
 #define MUTE_OP			0x43
-
 #define PLAY_OP			0x44
 #define STOP_OP			0x45
 #define PAUSE_OP		0x46
+#define RECORD_OP		0x47
 #define REWIND_OP		0x48
 #define FAST_FORWARD_OP		0x49
-#define NEXT_OP			0x4b
-#define PREV_OP			0x4c
+#define EJECT_OP		0x4a
+#define FORWARD_OP		0x4b
+#define BACKWARD_OP		0x4c
 
 static DBusConnection *connection = NULL;
 
 static GSList *servers = NULL;
-static GSList *sessions = NULL;
-
-typedef enum {
-	AVCTP_STATE_DISCONNECTED = 0,
-	AVCTP_STATE_CONNECTING,
-	AVCTP_STATE_CONNECTED
-} avctp_state_t;
 
 #if __BYTE_ORDER == __LITTLE_ENDIAN
 
@@ -116,6 +115,7 @@ struct avctp_header {
 	uint8_t transaction:4;
 	uint16_t pid;
 } __attribute__ ((packed));
+#define AVCTP_HEADER_LENGTH 3
 
 struct avrcp_header {
 	uint8_t code:4;
@@ -124,6 +124,7 @@ struct avrcp_header {
 	uint8_t subunit_type:5;
 	uint8_t opcode;
 } __attribute__ ((packed));
+#define AVRCP_HEADER_LENGTH 3
 
 #elif __BYTE_ORDER == __BIG_ENDIAN
 
@@ -134,6 +135,7 @@ struct avctp_header {
 	uint8_t ipid:1;
 	uint16_t pid;
 } __attribute__ ((packed));
+#define AVCTP_HEADER_LENGTH 3
 
 struct avrcp_header {
 	uint8_t _hdr0:4;
@@ -142,11 +144,18 @@ struct avrcp_header {
 	uint8_t subunit_id:3;
 	uint8_t opcode;
 } __attribute__ ((packed));
+#define AVRCP_HEADER_LENGTH 3
 
 #else
 #error "Unknown byte order"
 #endif
 
+struct avctp_state_callback {
+	avctp_state_cb cb;
+	void *user_data;
+	unsigned int id;
+};
+
 struct avctp_server {
 	bdaddr_t src;
 	GIOChannel *io;
@@ -154,26 +163,22 @@ struct avctp_server {
 	uint32_t ct_record_id;
 };
 
-struct avctp {
+struct control {
 	struct audio_device *dev;
 
 	avctp_state_t state;
 
-	bdaddr_t src;
-	bdaddr_t dst;
-
 	int uinput;
 
-	int sock;
-
-	guint io;
+	GIOChannel *io;
+	guint io_id;
 
 	uint16_t mtu;
+
+	gboolean target;
 };
 
-struct control {
-	struct avctp *session;
-};
+static GSList *avctp_callbacks = NULL;
 
 static sdp_record_t *avrcp_ct_record()
 {
@@ -301,45 +306,6 @@ static sdp_record_t *avrcp_tg_record()
 	return record;
 }
 
-static struct avctp *find_session(const bdaddr_t *src, const bdaddr_t *dst)
-{
-	GSList *l;
-
-	for (l = sessions; l != NULL; l = g_slist_next(l)) {
-		struct avctp *s = l->data;
-
-		if (bacmp(src, &s->src) || bacmp(dst, &s->dst))
-			continue;
-
-		return s;
-	}
-
-	return NULL;
-}
-
-static struct avctp *avctp_get(const bdaddr_t *src, const bdaddr_t *dst)
-{
-	struct avctp *session;
-
-	assert(src != NULL);
-	assert(dst != NULL);
-
-	session = find_session(src, dst);
-	if (session)
-		return session;
-
-	session = g_new0(struct avctp, 1);
-
-	session->uinput = -1;
-	session->sock = -1;
-	bacpy(&session->src, src);
-	bacpy(&session->dst, dst);
-
-	sessions = g_slist_append(sessions, session);
-
-	return session;
-}
-
 static int send_event(int fd, uint16_t type, uint16_t code, int32_t value)
 {
 	struct uinput_event event;
@@ -361,7 +327,7 @@ static void send_key(int fd, uint16_t key, int pressed)
 	send_event(fd, EV_SYN, SYN_REPORT, 0);
 }
 
-static void handle_panel_passthrough(struct avctp *session,
+static void handle_panel_passthrough(struct control *control,
 					const unsigned char *operands,
 					int operand_count)
 {
@@ -382,31 +348,31 @@ static void handle_panel_passthrough(struct avctp *session,
 	switch (operands[0] & 0x7F) {
 	case PLAY_OP:
 		debug("AVRCP: PLAY %s", status);
-		send_key(session->uinput, KEY_PLAYPAUSE, pressed);
+		send_key(control->uinput, KEY_PLAYCD, pressed);
 		break;
 	case STOP_OP:
 		debug("AVRCP: STOP %s", status);
-		send_key(session->uinput, KEY_STOP, pressed);
+		send_key(control->uinput, KEY_STOPCD, pressed);
 		break;
 	case PAUSE_OP:
 		debug("AVRCP: PAUSE %s", status);
-		send_key(session->uinput, KEY_PLAYPAUSE, pressed);
+		send_key(control->uinput, KEY_PAUSECD, pressed);
 		break;
-	case NEXT_OP:
-		debug("AVRCP: NEXT %s", status);
-		send_key(session->uinput, KEY_NEXTSONG, pressed);
+	case FORWARD_OP:
+		debug("AVRCP: FORWARD %s", status);
+		send_key(control->uinput, KEY_NEXTSONG, pressed);
 		break;
-	case PREV_OP:
-		debug("AVRCP: PREV %s", status);
-		send_key(session->uinput, KEY_PREVIOUSSONG, pressed);
+	case BACKWARD_OP:
+		debug("AVRCP: BACKWARD %s", status);
+		send_key(control->uinput, KEY_PREVIOUSSONG, pressed);
 		break;
 	case REWIND_OP:
 		debug("AVRCP: REWIND %s", status);
-		send_key(session->uinput, KEY_REWIND, pressed);
+		send_key(control->uinput, KEY_REWIND, pressed);
 		break;
 	case FAST_FORWARD_OP:
 		debug("AVRCP: FAST FORWARD %s", status);
-		send_key(session->uinput, KEY_FORWARD, pressed);
+		send_key(control->uinput, KEY_FASTFORWARD, pressed);
 		break;
 	default:
 		debug("AVRCP: unknown button 0x%02X %s", operands[0] & 0x7F, status);
@@ -414,53 +380,96 @@ static void handle_panel_passthrough(struct avctp *session,
 	}
 }
 
-static void avctp_unref(struct avctp *session)
+static void avctp_disconnected(struct audio_device *dev)
 {
-	sessions = g_slist_remove(sessions, session);
+	struct control *control = dev->control;
 
-	if (session->state == AVCTP_STATE_CONNECTED) {
-		gboolean value = FALSE;
-		g_dbus_emit_signal(session->dev->conn, session->dev->path,
-					AUDIO_CONTROL_INTERFACE,
-					"Disconnected", DBUS_TYPE_INVALID);
-		emit_property_changed(session->dev->conn, session->dev->path,
-					AUDIO_CONTROL_INTERFACE, "Connected",
-					DBUS_TYPE_BOOLEAN, &value);
+	if (!control)
+		return;
+
+	if (control->io) {
+		g_io_channel_shutdown(control->io, TRUE, NULL);
+		g_io_channel_unref(control->io);
+		control->io = NULL;
 	}
 
-	if (session->sock >= 0)
-		close(session->sock);
-	if (session->io)
-		g_source_remove(session->io);
-
-	if (session->dev)
-		session->dev->control->session = NULL;
-
-	if (session->uinput >= 0) {
-		ioctl(session->uinput, UI_DEV_DESTROY);
-		close(session->uinput);
+	if (control->io_id) {
+		g_source_remove(control->io_id);
+		control->io_id = 0;
 	}
 
-	g_free(session);
+	if (control->uinput >= 0) {
+		ioctl(control->uinput, UI_DEV_DESTROY);
+		close(control->uinput);
+		control->uinput = -1;
+	}
 }
 
-static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
+static void avctp_set_state(struct control *control, avctp_state_t new_state)
+{
+	GSList *l;
+	struct audio_device *dev = control->dev;
+	avdtp_session_state_t old_state = control->state;
+	gboolean value;
+
+	switch (new_state) {
+	case AVCTP_STATE_DISCONNECTED:
+		avctp_disconnected(control->dev);
+
+		if (old_state != AVCTP_STATE_CONNECTED)
+			break;
+
+		value = FALSE;
+		g_dbus_emit_signal(dev->conn, dev->path,
+					AUDIO_CONTROL_INTERFACE,
+					"Disconnected", DBUS_TYPE_INVALID);
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_CONTROL_INTERFACE, "Connected",
+					DBUS_TYPE_BOOLEAN, &value);
+		break;
+	case AVCTP_STATE_CONNECTING:
+		break;
+	case AVCTP_STATE_CONNECTED:
+		value = TRUE;
+		g_dbus_emit_signal(control->dev->conn, control->dev->path,
+				AUDIO_CONTROL_INTERFACE, "Connected",
+				DBUS_TYPE_INVALID);
+		emit_property_changed(control->dev->conn, control->dev->path,
+				AUDIO_CONTROL_INTERFACE, "Connected",
+				DBUS_TYPE_BOOLEAN, &value);
+		break;
+	default:
+		error("Invalid AVCTP state %d", new_state);
+		return;
+	}
+
+	control->state = new_state;
+
+	for (l = avctp_callbacks; l != NULL; l = l->next) {
+		struct avctp_state_callback *cb = l->data;
+		cb->cb(control->dev, old_state, new_state, cb->user_data);
+	}
+}
+
+static gboolean control_cb(GIOChannel *chan, GIOCondition cond,
 				gpointer data)
 {
-	struct avctp *session = data;
+	struct control *control = data;
 	unsigned char buf[1024], *operands;
 	struct avctp_header *avctp;
 	struct avrcp_header *avrcp;
-	int ret, packet_size, operand_count;
+	int ret, packet_size, operand_count, sock;
 
 	if (!(cond | G_IO_IN))
 		goto failed;
 
-	ret = read(session->sock, buf, sizeof(buf));
+	sock = g_io_channel_unix_get_fd(control->io);
+
+	ret = read(sock, buf, sizeof(buf));
 	if (ret <= 0)
 		goto failed;
 
-	debug("Got %d bytes of data for AVCTP session %p", ret, session);
+	debug("Got %d bytes of data for AVCTP session %p", ret, control);
 
 	if ((unsigned int) ret < sizeof(struct avctp_header)) {
 		error("Too small AVCTP packet");
@@ -495,21 +504,21 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 			avrcp->code, avrcp->subunit_type, avrcp->subunit_id,
 			avrcp->opcode, operand_count);
 
-	if (avctp->packet_type == AVCTP_PACKET_SINGLE &&
-			avctp->cr == AVCTP_COMMAND &&
-			avctp->pid == htons(AV_REMOTE_SVCLASS_ID) &&
+	if (avctp->packet_type != AVCTP_PACKET_SINGLE) {
+		avctp->cr = AVCTP_RESPONSE;
+		avrcp->code = CTYPE_NOT_IMPLEMENTED;
+	} else if (avctp->pid != htons(AV_REMOTE_SVCLASS_ID)) {
+		avctp->ipid = 1;
+		avctp->cr = AVCTP_RESPONSE;
+		avrcp->code = CTYPE_REJECTED;
+	} else if (avctp->cr == AVCTP_COMMAND &&
 			avrcp->code == CTYPE_CONTROL &&
 			avrcp->subunit_type == SUBUNIT_PANEL &&
 			avrcp->opcode == OP_PASSTHROUGH) {
-		handle_panel_passthrough(session, operands, operand_count);
+		handle_panel_passthrough(control, operands, operand_count);
 		avctp->cr = AVCTP_RESPONSE;
 		avrcp->code = CTYPE_ACCEPTED;
-		ret = write(session->sock, buf, packet_size);
-	}
-
-	if (avctp->packet_type == AVCTP_PACKET_SINGLE &&
-			avctp->cr == AVCTP_COMMAND &&
-			avctp->pid == htons(AV_REMOTE_SVCLASS_ID) &&
+	} else if (avctp->cr == AVCTP_COMMAND &&
 			avrcp->code == CTYPE_STATUS &&
 			(avrcp->opcode == OP_UNITINFO
 			|| avrcp->opcode == OP_SUBUNITINFO)) {
@@ -517,14 +526,17 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 		avrcp->code = CTYPE_STABLE;
 		debug("reply to %s", avrcp->opcode == OP_UNITINFO ?
 				"OP_UNITINFO" : "OP_SUBUNITINFO");
-		ret = write(session->sock, buf, packet_size);
+	} else {
+		avctp->cr = AVCTP_RESPONSE;
+		avrcp->code = CTYPE_REJECTED;
 	}
+	ret = write(sock, buf, packet_size);
 
 	return TRUE;
 
 failed:
-	debug("AVCTP session %p got disconnected", session);
-	avctp_unref(session);
+	debug("AVCTP session %p got disconnected", control);
+	avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
 	return FALSE;
 }
 
@@ -571,11 +583,11 @@ static int uinput_create(char *name)
 	ioctl(fd, UI_SET_EVBIT, EV_SYN);
 
 	ioctl(fd, UI_SET_KEYBIT, KEY_PLAYPAUSE);
-	ioctl(fd, UI_SET_KEYBIT, KEY_STOP);
+	ioctl(fd, UI_SET_KEYBIT, KEY_STOPCD);
 	ioctl(fd, UI_SET_KEYBIT, KEY_NEXTSONG);
 	ioctl(fd, UI_SET_KEYBIT, KEY_PREVIOUSSONG);
 	ioctl(fd, UI_SET_KEYBIT, KEY_REWIND);
-	ioctl(fd, UI_SET_KEYBIT, KEY_FORWARD);
+	ioctl(fd, UI_SET_KEYBIT, KEY_FASTFORWARD);
 
 	if (ioctl(fd, UI_DEV_CREATE, NULL) < 0) {
 		err = errno;
@@ -589,253 +601,179 @@ static int uinput_create(char *name)
 	return fd;
 }
 
-static void init_uinput(struct avctp *session)
+static void init_uinput(struct control *control)
 {
 	char address[18];
 
-	ba2str(&session->dst, address);
+	ba2str(&control->dev->dst, address);
 
-	session->uinput = uinput_create(address);
-	if (session->uinput < 0)
+	control->uinput = uinput_create(address);
+	if (control->uinput < 0)
 		error("AVRCP: failed to init uinput for %s", address);
 	else
 		debug("AVRCP: uinput initialized for %s", address);
 }
 
-static gboolean avctp_connect_session(struct avctp *session)
-{
-	GIOChannel *io;
-	gboolean value;
-
-	session->dev = manager_find_device(&session->dst, NULL, FALSE);
-	if (!session->dev) {
-		error("Connecting audio device not known (SDP not completed)");
-		return FALSE;
-	}
-
-	session->state = AVCTP_STATE_CONNECTED;
-
-	session->dev->control->session = session;
-
-	init_uinput(session);
-
-	value = TRUE;
-	g_dbus_emit_signal(session->dev->conn, session->dev->path,
-				AUDIO_CONTROL_INTERFACE, "Connected",
-				DBUS_TYPE_INVALID);
-	emit_property_changed(session->dev->conn, session->dev->path,
-				AUDIO_CONTROL_INTERFACE, "Connected",
-				DBUS_TYPE_BOOLEAN, &value);
-
-	if (session->io)
-		g_source_remove(session->io);
-
-	io = g_io_channel_unix_new(session->sock);
-	session->io = g_io_add_watch(io,
-			G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-			(GIOFunc) session_cb, session);
-	g_io_channel_unref(io);
-
-	return TRUE;
-}
-
-static void auth_cb(DBusError *derr, void *user_data)
-{
-	struct avctp *session = user_data;
-
-	if (derr && dbus_error_is_set(derr)) {
-		error("Access denied: %s", derr->message);
-
-		avctp_unref(session);
-		return;
-	}
-
-	avctp_connect_session(session);
-}
-
-static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-				const bdaddr_t *dst, gpointer data)
-{
-	socklen_t size;
-	struct l2cap_options l2o;
-	struct avctp *session;
-	GIOCondition flags = G_IO_ERR | G_IO_HUP | G_IO_NVAL;
-	struct audio_device *dev;
-	char address[18];
-
-	if (err < 0) {
-		error("AVCTP server socket: %s (%d)", strerror(-err), -err);
-		return;
-	}
-
-	session = avctp_get(src, dst);
-
-	if (!session) {
-		error("Unable to create new AVCTP session");
-		goto drop;
-	}
-
-	ba2str(dst, address);
-
-	if (session->sock >= 0) {
-		error("Refusing unexpected connect from %s", address);
-		goto drop;
-	}
-
-	dev = manager_get_device(src, dst);
-	if (!dev) {
-		error("Unable to get audio device object for %s", address);
-		goto drop;
-	}
-
-	if (!dev->control)
-		dev->control = control_init(dev);
-
-	device_remove_control_timer(dev);
-
-	session->state = AVCTP_STATE_CONNECTING;
-	session->sock = g_io_channel_unix_get_fd(chan);
-
-	memset(&l2o, 0, sizeof(l2o));
-	size = sizeof(l2o);
-	if (getsockopt(session->sock, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &size) < 0) {
-		err = errno;
-		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
-				err);
-		goto drop;
-	}
-
-	session->mtu = l2o.imtu;
-
-	session->io = g_io_add_watch(chan, flags, (GIOFunc) session_cb,
-				session);
-	g_io_channel_unref(chan);
-
-	if (avdtp_is_connected(src, dst))
-		goto proceed;
-
-	err = btd_request_authorization(src, dst, AVRCP_TARGET_UUID,
-				auth_cb, session);
-	if (err < 0)
-		goto drop;
-
-	return;
-
-proceed:
-	if (!avctp_connect_session(session))
-		goto drop;
-
-	return;
-
-drop:
-	g_io_channel_close(chan);
-	avctp_unref(session);
-}
-
-static GIOChannel *avctp_server_socket(const bdaddr_t *src, gboolean master)
-{
-	GIOChannel *io;
-	int lm = L2CAP_LM_AUTH | L2CAP_LM_ENCRYPT;
-
-	if (master)
-		lm |= L2CAP_LM_MASTER;
-
-	io = bt_l2cap_listen(src, AVCTP_PSM, 0, lm, avctp_server_cb,
-				NULL);
-	if (!io) {
-		error("Unable to allocate new io channel");
-		return NULL;
-	}
-
-	return io;
-}
-
-static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer data)
+static void avctp_connect_cb(GIOChannel *chan, GError *err, gpointer data)
 {
 	struct control *control = data;
-	struct avctp *session = control->session;
-	struct l2cap_options l2o;
-	socklen_t len;
-	int sk;
 	char address[18];
+	uint16_t imtu;
+	GError *gerr = NULL;
 
-	if (!session) {
-		debug("avctp_connect_cb: session removed while connecting");
-		g_io_channel_close(chan);
-		g_io_channel_unref(chan);
+	if (err) {
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
+		error("%s", err->message);
 		return;
 	}
 
-	ba2str(&session->dst, address);
-
-	if (err < 0) {
-		avctp_unref(session);
-		error("AVCTP connect(%s): %s (%d)", address, strerror(-err),
-				-err);
+	bt_io_get(chan, BT_IO_L2CAP, &gerr,
+			BT_IO_OPT_DEST, &address,
+			BT_IO_OPT_IMTU, &imtu,
+			BT_IO_OPT_INVALID);
+	if (gerr) {
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
+		error("%s", gerr->message);
+		g_error_free(gerr);
 		return;
 	}
 
 	debug("AVCTP: connected to %s", address);
 
-	g_io_channel_set_close_on_unref(chan, FALSE);
-	sk = g_io_channel_unix_get_fd(chan);
-	session->sock = sk;
+	if (!control->io)
+		control->io = g_io_channel_ref(chan);
 
-	memset(&l2o, 0, sizeof(l2o));
-	len = sizeof(l2o);
-	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &len) < 0) {
-		err = errno;
-		g_io_channel_unref(chan);
-		avctp_unref(session);
-		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
-				err);
+	init_uinput(control);
+
+	avctp_set_state(control, AVCTP_STATE_CONNECTED);
+	control->mtu = imtu;
+	control->io_id = g_io_add_watch(chan,
+				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+				(GIOFunc) control_cb, control);
+}
+
+static void auth_cb(DBusError *derr, void *user_data)
+{
+	struct control *control = user_data;
+	GError *err = NULL;
+
+	if (derr && dbus_error_is_set(derr)) {
+		error("Access denied: %s", derr->message);
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
 		return;
 	}
 
-	init_uinput(session);
+	if (!bt_io_accept(control->io, avctp_connect_cb, control,
+								NULL, &err)) {
+		error("bt_io_accept: %s", err->message);
+		g_error_free(err);
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
+	}
+}
 
-	g_dbus_emit_signal(session->dev->conn, session->dev->path,
-				AUDIO_CONTROL_INTERFACE, "Connected",
-				DBUS_TYPE_INVALID);
+static void avctp_confirm_cb(GIOChannel *chan, gpointer data)
+{
+	struct control *control = NULL;
+	struct audio_device *dev;
+	char address[18];
+	bdaddr_t src, dst;
+	GError *err = NULL;
 
-	session->state = AVCTP_STATE_CONNECTED;
-	session->mtu = l2o.imtu;
-	session->io = g_io_add_watch(chan,
-				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-				(GIOFunc) session_cb, session);
-	g_io_channel_unref(chan);
+	bt_io_get(chan, BT_IO_L2CAP, &err,
+			BT_IO_OPT_SOURCE_BDADDR, &src,
+			BT_IO_OPT_DEST_BDADDR, &dst,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		g_io_channel_shutdown(chan, TRUE, NULL);
+		return;
+	}
+
+	dev = manager_get_device(&src, &dst, TRUE);
+	if (!dev) {
+		error("Unable to get audio device object for %s", address);
+		goto drop;
+	}
+
+	if (!dev->control) {
+		btd_device_add_uuid(dev->btd_dev, AVRCP_REMOTE_UUID);
+		if (!dev->control)
+			goto drop;
+	}
+
+	control = dev->control;
+
+	if (control->io) {
+		error("Refusing unexpected connect from %s", address);
+		goto drop;
+	}
+
+	avctp_set_state(control, AVCTP_STATE_CONNECTING);
+	control->io = g_io_channel_ref(chan);
+
+	if (avdtp_is_connected(&src, &dst)) {
+		if (!bt_io_accept(chan, avctp_connect_cb, dev->control,
+								NULL, NULL))
+			goto drop;
+	} else if (btd_request_authorization(&src, &dst,
+				AVRCP_TARGET_UUID, auth_cb, dev->control) < 0)
+		goto drop;
+
+	return;
+
+drop:
+	if (!control || !control->io)
+		g_io_channel_shutdown(chan, TRUE, NULL);
+	if (control)
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
+}
+
+static GIOChannel *avctp_server_socket(const bdaddr_t *src, gboolean master)
+{
+	GError *err = NULL;
+	GIOChannel *io;
+
+	io = bt_io_listen(BT_IO_L2CAP, NULL, avctp_confirm_cb, NULL,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, src,
+				BT_IO_OPT_PSM, AVCTP_PSM,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_MEDIUM,
+				BT_IO_OPT_MASTER, master,
+				BT_IO_OPT_INVALID);
+	if (!io) {
+		error("%s", err->message);
+		g_error_free(err);
+	}
+
+	return io;
 }
 
 gboolean avrcp_connect(struct audio_device *dev)
 {
 	struct control *control = dev->control;
-	struct avctp *session;
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
-	if (control->session)
+	if (control->state > AVCTP_STATE_DISCONNECTED)
 		return TRUE;
 
-	session = avctp_get(&dev->src, &dev->dst);
-	if (!session) {
-		error("Unable to create new AVCTP session");
+	avctp_set_state(control, AVCTP_STATE_CONNECTING);
+
+	io = bt_io_connect(BT_IO_L2CAP, avctp_connect_cb, control, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_PSM, AVCTP_PSM,
+				BT_IO_OPT_INVALID);
+	if (err) {
+		avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
+		error("%s", err->message);
+		g_error_free(err);
 		return FALSE;
 	}
 
-	device_remove_control_timer(dev);
-
-	session->dev = dev;
-	session->state = AVCTP_STATE_CONNECTING;
-
-	err = bt_l2cap_connect(&dev->src, &dev->dst, AVCTP_PSM, 0,
-				avctp_connect_cb, control);
-	if (err < 0) {
-		avctp_unref(session);
-		error("Connect failed. %s(%d)", strerror(-err), -err);
-		return FALSE;
-	}
-
-	control->session = session;
+	control->io = io;
 
 	return TRUE;
 }
@@ -843,13 +781,11 @@ gboolean avrcp_connect(struct audio_device *dev)
 void avrcp_disconnect(struct audio_device *dev)
 {
 	struct control *control = dev->control;
-	struct avctp *session = control->session;
 
-	if (!session)
+	if (!(control && control->io))
 		return;
 
-	avctp_unref(session);
-	control->session = NULL;
+	avctp_set_state(control, AVCTP_STATE_DISCONNECTED);
 }
 
 int avrcp_register(DBusConnection *conn, const bdaddr_t *src, GKeyFile *config)
@@ -944,7 +880,7 @@ void avrcp_unregister(const bdaddr_t *src)
 	remove_record_from_server(server->ct_record_id);
 	remove_record_from_server(server->tg_record_id);
 
-	g_io_channel_close(server->io);
+	g_io_channel_shutdown(server->io, TRUE, NULL);
 	g_io_channel_unref(server->io);
 	g_free(server);
 
@@ -968,7 +904,7 @@ static DBusMessage *control_is_connected(DBusConnection *conn,
 	if (!reply)
 		return NULL;
 
-	connected = (control->session != NULL);
+	connected = (control->state == AVCTP_STATE_CONNECTED);
 
 	dbus_message_append_args(reply, DBUS_TYPE_BOOLEAN, &connected,
 					DBUS_TYPE_INVALID);
@@ -976,6 +912,100 @@ static DBusMessage *control_is_connected(DBusConnection *conn,
 	return reply;
 }
 
+static int avctp_send_passthrough(struct control *control, uint8_t op)
+{
+	unsigned char buf[AVCTP_HEADER_LENGTH + AVRCP_HEADER_LENGTH + 2];
+	struct avctp_header *avctp = (void *) buf;
+	struct avrcp_header *avrcp = (void *) &buf[AVCTP_HEADER_LENGTH];
+	uint8_t *operands = &buf[AVCTP_HEADER_LENGTH + AVRCP_HEADER_LENGTH];
+	int err, sk = g_io_channel_unix_get_fd(control->io);
+	static uint8_t transaction = 0;
+
+	memset(buf, 0, sizeof(buf));
+
+	avctp->transaction = transaction++;
+	avctp->packet_type = AVCTP_PACKET_SINGLE;
+	avctp->cr = AVCTP_COMMAND;
+	avctp->pid = htons(AV_REMOTE_SVCLASS_ID);
+
+	avrcp->code = CTYPE_CONTROL;
+	avrcp->subunit_type = SUBUNIT_PANEL;
+	avrcp->opcode = OP_PASSTHROUGH;
+
+	operands[0] = op & 0x7f;
+	operands[1] = 0;
+
+	err = write(sk, buf, sizeof(buf));
+	if (err < 0)
+		return err;
+
+	/* Button release */
+	avctp->transaction = transaction++;
+	operands[0] |= 0x80;
+
+	return write(sk, buf, sizeof(buf));
+}
+
+static DBusMessage *volume_up(DBusConnection *conn, DBusMessage *msg,
+								void *data)
+{
+	struct audio_device *device = data;
+	struct control *control = device->control;
+	DBusMessage *reply;
+	int err;
+
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return NULL;
+
+	if (control->state != AVCTP_STATE_CONNECTED)
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".NotConnected",
+					"Device not Connected");
+
+	if (!control->target)
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".NotSupported",
+					"AVRCP Target role not supported");
+
+	err = avctp_send_passthrough(control, VOL_UP_OP);
+	if (err < 0)
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+							strerror(-err));
+
+	return dbus_message_new_method_return(msg);
+}
+
+static DBusMessage *volume_down(DBusConnection *conn, DBusMessage *msg,
+								void *data)
+{
+	struct audio_device *device = data;
+	struct control *control = device->control;
+	DBusMessage *reply;
+	int err;
+
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return NULL;
+
+	if (control->state != AVCTP_STATE_CONNECTED)
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".NotConnected",
+					"Device not Connected");
+
+	if (!control->target)
+		return g_dbus_create_error(msg,
+					ERROR_INTERFACE ".NotSupported",
+					"AVRCP Target role not supported");
+
+	err = avctp_send_passthrough(control, VOL_DOWN_OP);
+	if (err < 0)
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+							strerror(-err));
+
+	return dbus_message_new_method_return(msg);
+}
+
 static DBusMessage *control_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -997,7 +1027,7 @@ static DBusMessage *control_get_properties(DBusConnection *conn,
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
 	/* Connected */
-	value = (device->control->session != NULL);
+	value = (device->control->state == AVCTP_STATE_CONNECTED);
 	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &value);
 
 	dbus_message_iter_close_container(&iter, &dict);
@@ -1009,6 +1039,8 @@ static GDBusMethodTable control_methods[] = {
 	{ "IsConnected",	"",	"b",	control_is_connected,
 						G_DBUS_METHOD_FLAG_DEPRECATED },
 	{ "GetProperties",	"",	"a{sv}",control_get_properties },
+	{ "VolumeUp",		"",	"",	volume_up },
+	{ "VolumeDown",		"",	"",	volume_down },
 	{ NULL, NULL, NULL, NULL }
 };
 
@@ -1019,25 +1051,19 @@ static GDBusSignalTable control_signals[] = {
 	{ NULL, NULL }
 };
 
-static void control_free(struct audio_device *dev)
-{
-	struct control *control = dev->control;
-
-	if (control->session)
-		avctp_unref(control->session);
-
-	g_free(control);
-	dev->control = NULL;
-}
-
 static void path_unregister(void *data)
 {
 	struct audio_device *dev = data;
+	struct control *control = dev->control;
 
-	info("Unregistered interface %s on path %s",
+	debug("Unregistered interface %s on path %s",
 		AUDIO_CONTROL_INTERFACE, dev->path);
 
-	control_free(dev);
+	if (control->state != AVCTP_STATE_DISCONNECTED)
+		avctp_disconnected(dev);
+
+	g_free(control);
+	dev->control = NULL;
 }
 
 void control_unregister(struct audio_device *dev)
@@ -1046,27 +1072,75 @@ void control_unregister(struct audio_device *dev)
 		AUDIO_CONTROL_INTERFACE);
 }
 
-struct control *control_init(struct audio_device *dev)
+void control_update(struct audio_device *dev, uint16_t uuid16)
 {
+	struct control *control = dev->control;
+
+	if (uuid16 == AV_REMOTE_TARGET_SVCLASS_ID)
+		control->target = TRUE;
+}
+
+struct control *control_init(struct audio_device *dev, uint16_t uuid16)
+{
+	struct control *control;
+
 	if (!g_dbus_register_interface(dev->conn, dev->path,
 					AUDIO_CONTROL_INTERFACE,
 					control_methods, control_signals, NULL,
 					dev, path_unregister))
 		return NULL;
 
-	info("Registered interface %s on path %s",
+	debug("Registered interface %s on path %s",
 		AUDIO_CONTROL_INTERFACE, dev->path);
 
-	return g_new0(struct control, 1);
+	control = g_new0(struct control, 1);
+	control->dev = dev;
+	control->state = AVCTP_STATE_DISCONNECTED;
+	control->uinput = -1;
+
+	if (uuid16 == AV_REMOTE_TARGET_SVCLASS_ID)
+		control->target = TRUE;
+
+	return control;
 }
 
 gboolean control_is_active(struct audio_device *dev)
 {
 	struct control *control = dev->control;
 
-	if (control->session &&
-			control->session->state != AVCTP_STATE_DISCONNECTED)
+	if (control && control->state != AVCTP_STATE_DISCONNECTED)
 		return TRUE;
 
 	return FALSE;
 }
+
+unsigned int avctp_add_state_cb(avctp_state_cb cb, void *user_data)
+{
+	struct avctp_state_callback *state_cb;
+	static unsigned int id = 0;
+
+	state_cb = g_new(struct avctp_state_callback, 1);
+	state_cb->cb = cb;
+	state_cb->user_data = user_data;
+	state_cb->id = ++id;
+
+	avctp_callbacks = g_slist_append(avctp_callbacks, state_cb);
+
+	return state_cb->id;
+}
+
+gboolean avctp_remove_state_cb(unsigned int id)
+{
+	GSList *l;
+
+	for (l = avctp_callbacks; l != NULL; l = l->next) {
+		struct avctp_state_callback *cb = l->data;
+		if (cb && cb->id == id) {
+			avctp_callbacks = g_slist_remove(avctp_callbacks, cb);
+			g_free(cb);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
